from types import ModuleType
from typing import Optional, List

from django.utils.module_loading import module_has_submodule

try:
    from django.utils.module_loading import import_module

except ImportError:
    # Django <=1.9.0
    from django.utils.importlib import import_module


def import_app_module(app_name: str, module_name: str) -> Optional[ModuleType]:
    """Returns a module from a given app by its name.

    :param app_name:
    :param module_name:

    """
    name_split = app_name.split('.')
    if name_split[-1][0].isupper():  # Seems that we have app config class path here.
        app_name = '.'.join(name_split[:-2])

    module = import_module(app_name)

    try:
        sub_module = import_module(f'{app_name}.{module_name}')
        return sub_module

    except:

        # The same bubbling strategy as in autodiscover_modules().
        if module_has_submodule(module, module_name):  # Module is in a package.
            raise

        return None


def import_project_modules(module_name: str) -> List[ModuleType]:
    """Imports modules from registered apps using given module name
    and returns them as a list.

    :param module_name:

    """
    from django.conf import settings

    submodules = []
    for app in settings.INSTALLED_APPS:
        module = import_app_module(app, module_name)
        if module is not None:
            submodules.append(module)

    return submodules



from functools import partial
from os import environ
from typing import Optional

from django.conf import settings
from django.contrib.sites.shortcuts import get_current_site
from django.http import HttpRequest


class DomainGetter:

    __slots__ = ['domain']

    def __init__(self, domain: Optional[str]):
        self.domain = domain

    def get_host(self) -> Optional[str]:
        return self.domain


def get_site_url(request: HttpRequest = None) -> str:
    """Tries to get a site URL from environment and settings
    in the following order:

    1. (SITE_PROTO / SITE_SCHEME) + SITE_DOMAIN
    2. SITE_URL
    3. Django Sites contrib
    4. Request object

    :param request: Request object to deduce URL from.

    """
    env = partial(environ.get)
    settings_ = partial(getattr, settings)

    domain = None
    scheme = None
    url = None

    for src in (env, settings_):
        if url is None:
            url = src('SITE_URL', None)

        if domain is None:
            domain = src('SITE_DOMAIN', None)

        if scheme is None:
            scheme = src('SITE_PROTO', src('SITE_SCHEME', None))

    if domain is None and url is not None:
        scheme, domain = url.split('://')[:2]

    if domain is None:
        site = get_current_site(request or DomainGetter(domain))
        domain = site.domain

    if scheme is None and request:
        scheme = request.scheme

    if domain is None:
        domain = 'undefined-domain.local'

    if scheme is None:
        scheme = 'http'

    domain = domain.rstrip('/')

    return f'{scheme}://{domain}'



from django.db import models

from etc.admin import CustomModelPage


class MyPage1(CustomModelPage):

    title = 'Test page 1'
    my_field = models.CharField('some title', max_length=10)

    def save(self):
        self.bound_admin.message_warning(self.bound_request, f'test1:{self.my_field}')
        super().save()


class MyPage2(CustomModelPage):

    title = 'Second test page'
    my_another_field = models.TextField('put data here')

    def save(self):
        self.bound_admin.message_error(self.bound_request, f'test2:{self.my_another_field}')


MyPage1.register()
MyPage2.register()



import hashlib

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from django import template
from django.template.defaultfilters import safe
from django.contrib.auth import get_user_model

USER_MODEL = get_user_model()

register = template.Library()


def get_gravatar_url(obj, size=65, default='identicon'):
    """

    See http://ru.gravatar.com/site/implement/images/ for details on options.

    :param UserModel, str obj:
    :param int size:
    :param str default: 404, mm (mystery-man), identicon, monsterid, wavatar, retro, blank
    :return:
    """
    if isinstance(obj, USER_MODEL):
        email = obj.email or obj.username
    else:
        email = obj

    if email:
        return ('http://www.gravatar.com/avatar/%s/?%s' %
                (hashlib.md5(email.encode()).hexdigest(), urlencode({'size': size, 'd': default})))
    return ''


@register.simple_tag
def gravatar_get_url(obj, size=65, default='identicon'):
    """Returns Gravatar image URL for a given string or UserModel.

    Example:

        {% load gravatar %}
        {% gravatar_get_url user_model %}

    :param UserModel, str obj:
    :param int size:
    :param str default:
    :return:
    """
    return get_gravatar_url(obj, size=size, default=default)


@register.simple_tag
def gravatar_get_img(obj, size=65, default='identicon'):
    """Returns Gravatar image HTML tag for a given string or UserModel.

    Example:

        {% load gravatar %}
        {% gravatar_get_img user_model %}

    :param UserModel, str obj:
    :param int size:
    :param str default:
    :return:
    """
    url = get_gravatar_url(obj, size=size, default=default)
    if url:
        return safe('<img src="%s" class="gravatar">' % url)
    return ''



from sys import version_info
from django.db import models

from etc.toolbox import InheritedModel


if version_info.major != 2:

    from etc.toolbox import ChoicesEnumMixin, get_choices

    from enum import Enum, unique

    @unique
    class Role(ChoicesEnumMixin, Enum):

        APPLICANT = 0, 'Applicant', 'Description'
        ADMIN = 1, 'Administrator'
        MEMBER = 2

    class Variant(ChoicesEnumMixin, Enum):

        A = 'a'
        B = 'b'

    class MyChoiceModel(models.Model):

        role = models.PositiveIntegerField(
            'Role', choices=get_choices(Role), default=Role.MEMBER)

        variant = models.CharField(
            choices=get_choices(Variant), default=Variant.A, max_length='2')


class MyParentModel(models.Model):

    code = models.CharField('dummy', max_length=64)
    expired = models.BooleanField('Expired', help_text='dummy', default=False)

    class Meta:
        abstract = True


class MyChildModel1(InheritedModel, MyParentModel):

    class Fields:
        code = 'Secret code'
        expired = {'help_text': 'This code is expired.'}

    class Meta:
        verbose_name = 'Verb'
        verbose_name_plural = 'VerbPlural'


class MyChildModel2(InheritedModel, MyParentModel):

    class Fields:
        code = 'Non-secret code'



from .choices import choices_list, get_choices, ChoicesEnumMixin
from .importing import import_app_module, import_project_modules
from .forms import set_form_widgets_attrs
from .models import get_model_class_from_settings, get_model_class_from_string, InheritedModel
from .sites import get_site_url



from django.contrib import admin
from django.contrib import messages
from django.db import models
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.urls import path

if False:  # pragma: nocover
    from .models import CustomModelPage  # noqa


class EtcAdmin(admin.ModelAdmin):
    """Base etc admin."""

    def message_success(self, request: HttpRequest, msg: str):
        self.message_user(request, msg, messages.SUCCESS)

    def message_warning(self, request: HttpRequest, msg: str):
        self.message_user(request, msg, messages.WARNING)

    def message_error(self, request: HttpRequest, msg: str):
        self.message_user(request, msg, messages.ERROR)


class ReadonlyAdmin(EtcAdmin):
    """Read-only etc admin base class."""

    view_on_site: bool = False
    actions = None

    def has_add_permission(self, request: HttpRequest) -> bool:
        return False

    def has_delete_permission(self, request: HttpRequest, obj: models.Model = None) -> bool:
        return False

    def changeform_view(
        self,
        request: HttpRequest,
        object_id: int = None,
        form_url: str = '',
        extra_context: dict = None
    ) -> HttpResponse:

        extra_context = extra_context or {}
        extra_context.update({
            'show_save_and_continue': False,
            'show_save': False,
        })
        return super().changeform_view(request, object_id, extra_context=extra_context)


class CustomPageModelAdmin(ReadonlyAdmin):
    """Base for admin pages with contents based on custom models."""

    def get_urls(self) -> list:
        meta = self.model._meta
        patterns = [path(
            '',
            self.admin_site.admin_view(self.view_custom),
            name=f'{meta.app_label}_{meta.model_name}_changelist'
        )]
        return patterns

    def has_add_permission(self, request: HttpRequest) -> bool:
        return True

    def view_custom(self, request: HttpRequest) -> HttpResponse:
        context: dict = {
            'show_save_and_continue': False,
            'show_save_and_add_another': False,
            'title': self.model._meta.verbose_name,
        }
        return self._changeform_view(request, object_id=None, form_url='', extra_context=context)

    def response_add(self, request: HttpRequest, obj: 'CustomModelPage', post_url_continue=None):
        return HttpResponseRedirect(request.path)

    def save_model(self, request: HttpRequest, obj: 'CustomModelPage', form, change):
        obj.bound_request = request
        obj.bound_admin = self
        obj.save()



from os import environ
from sys import version_info

import pytest
from django import forms
from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.template.base import TemplateSyntaxError

from etc.templatetags.gravatar import gravatar_get_url, gravatar_get_img
from etc.templatetags.model_meta import model_meta_verbose_name, model_meta_verbose_name_plural
from etc.toolbox import set_form_widgets_attrs, choices_list, get_choices, get_site_url, get_model_class_from_string, \
    get_model_class_from_settings, import_app_module, import_project_modules


class MyForm(forms.Form):

    field1 = forms.CharField(label='f1', max_length=100)
    field2 = forms.CharField(label='f2', max_length=100)


from etc.tests.testapp.models import MyChildModel1, MyChildModel2


class TestInheritedModel:

    def test_texts(self):
        model1 = MyChildModel1()
        fields = {f.name: f for f in model1._meta.fields}
        assert fields['code'].verbose_name == 'Secret code'
        assert fields['expired'].help_text == 'This code is expired.'

        model1 = MyChildModel2()
        fields = {f.name: f for f in model1._meta.fields}
        assert fields['code'].verbose_name == 'Non-secret code'
        assert fields['expired'].help_text == 'dummy'


class TestModelMetaTemplateTags:

    def test_verbose_name_singular(self):
        m = MyChildModel1
        assert model_meta_verbose_name(m) == 'Verb'

    def test_verbose_name_plural(self):
        m = MyChildModel1
        assert model_meta_verbose_name_plural(m) == 'VerbPlural'


class TestModelFieldTemplateTags:

    def test_model_field_x(self, template_render_tag, template_context, settings):

        def check(tag, val, fieldname):

            context = template_context({'model': User()})
            result = template_render_tag('model_field', '%s from %s' % (tag, fieldname), context)
            assert val in result
    
            #
            # `from` missing
            with pytest.raises(TemplateSyntaxError):
                template_render_tag('model_field', '%s %s' % (tag, fieldname))
    
            #
            # `as` clause
            context = template_context({'model': User(), 'field': fieldname.partition('.')[2]})
            result = template_render_tag('model_field', '%s from %s as a' % (tag, fieldname), context)
            assert result == ''
            assert val in context['a']
    
            # field in a variable
            result = template_render_tag('model_field', '%s from model.field' % tag, context)
            assert val in result
    
            #
            # Wrong model-field delimiter.
            result = template_render_tag('model_field', '%s from %s' % (tag, fieldname.replace('.', '-')), context)
            assert result == ''
    
            with settings(DEBUG=True):
                with pytest.raises(TemplateSyntaxError):
                    template_render_tag('model_field', '%s from %s' % (tag, fieldname.replace('.', '-')))
    
            #
            # No model in context.
            result = template_render_tag('model_field', '%s from %s' % (tag, fieldname))
            assert result == ''
    
            with settings(DEBUG=True):
                with pytest.raises(TemplateSyntaxError):
                    template_render_tag('model_field', '%s from %s' % (tag, fieldname))
    
            #
            # No field.
            result = template_render_tag('model_field', '%s from model.unknown' % tag, context)
            assert result == ''
    
            with settings(DEBUG=True):
                with pytest.raises(TemplateSyntaxError):
                    template_render_tag('model_field', '%s from model.unknown' % tag, context)

        check('model_field_verbose_name', 'first name', 'model.first_name')
        check('model_field_help_text', 'whether the user can log', 'model.is_staff')


class TestForm:

    def test_set_form_widgets_attrs(self):
        f = MyForm()

        d = {}
        d['class'] = 'clickable'
        d['data-a'] = lambda f: f.__class__.__name__

        set_form_widgets_attrs(f, d)
        output = f.as_p()

        assert output.count('data-a') == 2
        assert output.count('clickable') == 2


class TestGravatarTemplateTags:

    def test_verbose_get_url(self):
        u = User(username='idle')
        url = gravatar_get_url(u, 101, 'retro')
        assert 'http://www.gravatar.com/avatar/ec2f993aec2c27fc750119ab17b16cdb/' in url
        assert 'retro' in url
        assert '101' in url

        u = User(email='idle@sign.som')
        url = gravatar_get_url(u, 101, 'retro')
        assert 'http://www.gravatar.com/avatar/37e24208b31f2a8f1e0f84d4c93fdfb0/' in url
        assert 'retro' in url
        assert '101' in url

        url = gravatar_get_url('idle@sign.som', 101, 'retro')
        assert 'http://www.gravatar.com/avatar/37e24208b31f2a8f1e0f84d4c93fdfb0/' in url
        assert 'retro' in url
        assert '101' in url

        assert gravatar_get_url(None) == ''

    def test_verbose_get_img(self):
        u = User(username='idle')
        url = gravatar_get_img(u, 101, 'retro')
        assert 'http://www.gravatar.com/avatar/ec2f993aec2c27fc750119ab17b16cdb/' in url
        assert 'retro' in url
        assert 'retro' in url
        assert '<img src="' in url

        assert gravatar_get_img(None) == ''


class TestChoices:

    @pytest.mark.skipif(version_info.major == 2, reason='No Enum in Python 2')
    def test_enum(self):

        from etc.tests.testapp.models import Role, MyChoiceModel, Variant

        assert Role.APPLICANT is Role(0)
        assert Role.APPLICANT.value == 0
        assert Role.APPLICANT.title == 'Applicant'
        assert Role.APPLICANT.hint == 'Description'

        assert Role.ADMIN is Role(1)
        assert Role.ADMIN.value == 1
        assert Role.ADMIN.title == 'Administrator'
        assert Role.ADMIN.hint == ''

        assert Role.MEMBER is Role(2)
        assert Role.MEMBER.value == 2
        assert Role.MEMBER.title == 'Member'
        assert Role.MEMBER.hint == ''

        assert Role.get_title(2) == 'Member'
        assert Role.get_title(Role(2)) == 'Member'
        assert Role.get_title(Role.MEMBER) == 'Member'

        assert Role.get_hint(0) == 'Description'
        assert Role.get_hint(Role(0)) == 'Description'
        assert Role.get_hint(Role.APPLICANT) == 'Description'

        objects = MyChoiceModel.objects

        obj_1 = objects.create()
        obj_2 = objects.create(role=Role.ADMIN, variant=Variant.B)

        assert obj_1.role is Role.MEMBER
        assert obj_1.variant is Variant.A

        assert obj_1.id == objects.get(role=Role.MEMBER).id
        assert obj_2.id == objects.get(variant=Variant.B).id

    def test_choices(self):

        types_dict = choices_list(
            (1, 'T1'),
            (2, 'T2'),
        )

        assert len(types_dict) == 2
        assert types_dict[1] == 'T1'
        assert types_dict[2] == 'T2'

        assert list(types_dict.keys()) == [1, 2]

        ch = get_choices(types_dict)
        assert len(ch) == 2

        assert ch[0][0] == 1
        assert ch[0][1] == 'T1'

        assert ch[1][0] == 2
        assert ch[1][1] == 'T2'


class TestGetModelClass:

    def test_from_settings(self):
        attr_name = 'some'
        fake_settings = type('fake', (object,), {attr_name: 'auth.User'})

        assert get_model_class_from_settings(fake_settings, attr_name) is User

    def test_from_string(self):
        cl = get_model_class_from_string('auth.User')
        assert cl is User

        with pytest.raises(ImproperlyConfigured):
            get_model_class_from_string('some')

        with pytest.raises(ImproperlyConfigured):
            get_model_class_from_string('etc.InheritedModel')


class TestGetSiteUrl:

    def test_basic(self, settings):
        assert get_site_url() == 'http://example.com'

        with settings(SITE_PROTO='htt'):
            assert get_site_url() == 'htt://example.com'
            environ['SITE_PROTO'] = 'ttp'
            assert get_site_url() == 'ttp://example.com'

        environ['SITE_PROTO'] = 'https'
        assert get_site_url() == 'https://example.com'

        environ['SITE_SCHEME'] = 'ftp'
        assert get_site_url() == 'https://example.com'

        del environ['SITE_PROTO']
        assert get_site_url() == 'ftp://example.com'

        environ['SITE_URL'] = 'http://pythonz.net'
        assert get_site_url() == 'http://pythonz.net'

        environ['SITE_DOMAIN'] = 'mydomain.loc'
        assert get_site_url() == 'ftp://mydomain.loc'

        environ['SITE_URL'] = 'http://pythonz.net'
        assert get_site_url() == 'ftp://mydomain.loc'

        del environ['SITE_DOMAIN']
        del environ['SITE_URL']
        del environ['SITE_SCHEME']

        class FakeRequest:

            scheme = 'xyz'

            @classmethod
            def get_host(cls):
                return 'fake'

        assert get_site_url(request=FakeRequest) == 'xyz://example.com'

    def test_tempalte_tag(self, template_render_tag):
        url = 'http://pythonz.net'

        environ['SITE_URL'] = url
        result = template_render_tag('etc_misc', 'site_url')
        assert result == url


class TestImportModules:

    def test_import_app_module(self):
        m = import_app_module('etc', 'sites')
        assert hasattr(m, 'DomainGetter')

        with pytest.raises(ImportError):
            import_app_module('unknown', 'uknown')

        m = import_app_module('etc', 'uknown')
        assert m is None

        m = import_app_module('django.contrib.admin.apps.SimpleAdminConfig', '__init__')
        assert hasattr(m, 'site')

    def test_import_project_modules(self):
        m = import_project_modules('toolbox')

        assert len(m) == 1
        assert hasattr(m[0], 'get_site_url')


def test_include_formatted(request_client):
    result = request_client().get('/index/')
    assert result.content == b'\n<body>thisone\nstatic\n<sub>dynamic</sub>\ndefault\n</body>'



from django.forms import Form


def set_form_widgets_attrs(form: Form, attrs: dict):
    """Applies a given HTML attributes to each field widget of a given form.

    Example:

        set_form_widgets_attrs(my_form, {'class': 'clickable'})

    """
    for _, field in form.fields.items():
        attrs_ = dict(attrs)

        for name, val in attrs.items():
            if hasattr(val, '__call__'):
                attrs_[name] = val(field)

        field.widget.attrs = field.widget.build_attrs(attrs_)



from django import template
from django.conf import settings
from django.core.exceptions import FieldDoesNotExist
from django.core.paginator import Page
from django.db.models.query import QuerySet

register = template.Library()


@register.tag
def model_field_verbose_name(parser, token):
    """Returns model field verbose name.
    Two notations are acceptable:

        1. Two arguments:
           {% model_field_verbose_name from mymodel.myfield %}

           Is used to render verbose name of `myfield`.

        2. Four arguments:
           {% model_field_verbose_name from mymodel.myfield as myvar %}

           Is used to put `myfield` verbose name into `myvar` template variable.

    """
    return _get_model_field_attr('model_field_verbose_name', 'verbose_name', token)


@register.tag
def model_field_help_text(parser, token):
    """Returns model field help text.
    Two notations are acceptable:

        1. Two arguments:
           {% model_field_help_text from mymodel.myfield %}

           Is used to render verbose name of `myfield`.

        2. Four arguments:
           {% model_field_help_text from mymodel.myfield as myvar %}

           Is used to put `myfield` verbose name into `myvar` template variable.

    """
    return _get_model_field_attr('model_field_help_text', 'help_text', token)


def _get_model_field_attr(tag_name, attr_name, token):
    tokens = token.split_contents()
    tokens_num = len(tokens)

    if tokens_num not in (3, 5):
        raise template.TemplateSyntaxError(
            '`%(tag_name)s` tag requires two or four arguments. '
            'E.g.: {%% %(tag_name)s from model.field %%} or {%% %(tag_name)s from model.field as myvar %%}.'
            % {'tag_name': tag_name}
        )

    field = tokens[2]
    as_var = None

    tokens = tokens[3:]
    if len(tokens) >= 2 and tokens[-2] == 'as':
        as_var = tokens[-1]

    return FieldAttrNode(field, attr_name, tag_name, as_var)


class FieldAttrNode(template.Node):

    def __init__(self, field, attr_name, tag_name, as_var=None):
        self.tag_name = tag_name
        self.field = field
        self.attr_name = attr_name
        self.as_var = as_var

    def render(self, context):
        as_var = self.as_var

        def return_contents(contents):
            contents = f'{contents}'
            if as_var is not None:
                context[as_var] = contents
                return ''
            return contents

        def get_field(model, field_name):
            return model._meta.get_field(field_name)

        var_field = template.Variable(self.field)
        var_model = var_field.lookups[0]

        try:
            field_name = var_field.lookups[1]
        except IndexError:
            if settings.DEBUG:
                raise template.TemplateSyntaxError(
                    '`%s` template tag requires model.field notation but `%s` is given.' %
                    (self.tag_name, self.field)
                )
            return return_contents('')

        try:
            model = template.Variable(var_model).resolve(context)
        except template.VariableDoesNotExist:
            if settings.DEBUG:
                raise template.TemplateSyntaxError(
                    '`%s` template tag error: `%s` model is not found in context.' %
                    (self.tag_name, var_model)
                )
            return return_contents('')

        # Allow operations on homogeneous sets -- Pages and Query Sets.
        if isinstance(model, Page):
            model = model.object_list

        if isinstance(model, QuerySet):
            model = model.model

        fields_name_map = getattr(model._meta, '_name_map', None)
        if fields_name_map is None:
            fields_name_map = {f.attname: (f,) for f in model._meta.fields}

        try:

            try:
                # case #1: field name is literal
                model_field = get_field(model, field_name)

            except FieldDoesNotExist:
                # case #1: field name is a variable
                field_name = template.Variable(field_name).resolve(context)
                model_field = get_field(model, field_name)

            contents = getattr(model_field, self.attr_name)

        except (FieldDoesNotExist, template.VariableDoesNotExist):
            contents = ''
            if settings.DEBUG:
                raise template.TemplateSyntaxError(
                    '`%s` template tag error: `%s` field is not found in `%s` model. '
                    'Possible choices: %s.' % (
                        self.tag_name,
                        field_name,
                        model.__class__.__name__,
                        ', '.join(fields_name_map.keys())))

        return return_contents(contents)



from copy import deepcopy
from functools import partial

from django import template
from django.template import TemplateDoesNotExist
from django.template.base import Lexer, Parser
from django.template.loader_tags import do_include, Node

try:
    from django.template.loader_tags import construct_relative_path

    # To prevent AttributeError: 'Parser' object has no attribute 'origin'
    def construct_relative_path_(parser, name):
        return construct_relative_path(parser.origin.template_name, name)

except ImportError:
    construct_relative_path_ = lambda parser, name: name  # Sorry sub here for now.

from ..toolbox import get_site_url

get_lexer = partial(Lexer)
register = template.Library()


@register.simple_tag(takes_context=True)
def site_url(context):
    """Tries to get a site URL from environment and settings.

    See toolbox.get_site_url() for description.

    Example:

        {% load etc_misc %}
        {% site_url %}

    """
    return get_site_url(request=context.get('request', None))


class DynamicIncludeNode(Node):

    context_key = '__include_context'

    def __init__(self, template, *args, **kwargs):
        self.fallback = kwargs.pop('fallback', None)
        self.template = template
        self.extra_context = kwargs.pop('extra_context', {})
        self.isolated_context = kwargs.pop('isolated_context', False)
        super(DynamicIncludeNode, self).__init__(*args, **kwargs)

    def render_(self, tpl_new, context):

        template = deepcopy(self.template)  # Do not mess with global template for threadsafety.
        template.var = tpl_new  # Cheat a little

        # Below is implementation from Django 2.1 generic IncludeNode.

        template = template.resolve(context)
        # Does this quack like a Template?
        if not callable(getattr(template, 'render', None)):
            # If not, try the cache and get_template().
            template_name = template
            cache = context.render_context.dicts[0].setdefault(self, {})
            template = cache.get(template_name)
            if template is None:
                template = context.template.engine.get_template(template_name)
                cache[template_name] = template
        # Use the base.Template of a backends.django.Template.
        elif hasattr(template, 'template'):
            template = template.template
        values = {
            name: var.resolve(context)
            for name, var in self.extra_context.items()
        }
        if self.isolated_context:
            return template.render(context.new(values))
        with context.push(**values):
            return template.render(context)

    def render(self, context):
        render_ = self.render_

        try:
            return render_(
                tpl_new=Parser(get_lexer(self.template.var).tokenize()).parse().render(context),
                context=context)

        except TemplateDoesNotExist:
            fallback = self.fallback

            if not fallback:  # pragma: nocover
                raise

            return render_(tpl_new=fallback.var, context=context)


@register.tag('include_')
def include_(parser, token):
    """Similar to built-in ``include`` template tag, but allowing
    template variables to be used in template name and a fallback template,
    thus making the tag more dynamic.

    .. warning:: Requires Django 1.8+

    Example:

        {% load etc_misc %}
        {% include_ "sub_{{ postfix_var }}.html" fallback "default.html" %}

    """
    bits = token.split_contents()

    dynamic = False

    # We fallback to built-in `include` if a template name contains no variables.
    if len(bits) >= 2:
        dynamic = '{{' in bits[1]

        if dynamic:
            fallback = None
            bits_new = []

            for bit in bits:

                if fallback is True:
                    # This bit is a `fallback` argument.
                    fallback = bit
                    continue

                if bit == 'fallback':
                    fallback = True

                else:
                    bits_new.append(bit)

            if fallback:
                fallback = parser.compile_filter(construct_relative_path_(parser, fallback))

            token.contents = ' '.join(bits_new)

    token.contents = token.contents.replace('include_', 'include')
    include_node = do_include(parser, token)

    if dynamic:
        # swap simple include with dynamic
        include_node = DynamicIncludeNode(
            include_node.template,
            extra_context=include_node.extra_context,
            isolated_context=include_node.isolated_context,
            fallback=fallback or None,
        )

    return include_node



from types import ModuleType
from typing import Type

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.db.models.base import ModelBase, Model


class InheritedModelMetaclass(ModelBase):

    def __new__(cls, name, bases, attrs):
        cl = super(InheritedModelMetaclass, cls).__new__(cls, name, bases, attrs)

        texts_marker = '_texts_applied'

        if not getattr(cl, texts_marker, False):
            try:
                names_map = {f.name: f for f in cl._meta.fields}

            except AttributeError:
                return cl

            fields_cl = getattr(cl, 'Fields', None)

            if fields_cl is not None:
                for attr_name, val in fields_cl.__dict__.items():
                    if attr_name.startswith('_'):
                        continue

                    if attr_name not in names_map:
                        continue

                    field = names_map[attr_name]

                    if not isinstance(val, dict):
                        val = {'verbose_name': val}

                    for field_attr, field_val in val.items():
                        setattr(field, field_attr, field_val)

            setattr(cl, texts_marker, True)

        return cl


class InheritedModel(metaclass=InheritedModelMetaclass):
    """Mix in this class into target model (inherit from it) and define `Fields` class inside it
    to be able to customize field attributes (e.g. texts) of a base-parent model.


    Example:

        from etc.models import InheritedModel


        class MyParentModel(models.Model):

            code = models.CharField('dummy', max_length=64)
            expired = models.BooleanField('Expired', help_text='dummy')

            class Meta:
                abstract = True


        class MyChildModel1(InheritedModel, MyParentModel):  # NOTE: InheritedModel must go first.

            time_created = models.DateTimeField('Date created', auto_now_add=True)

            class Fields:  # Defining a class with fields custom fields data.
                code = 'Secret code'  # This is treated as verbose_name.
                expired = {'help_text': 'This code is expired.'}


        class MyChildModel2(InheritedModel, MyParentModel):

            code = models.CharField('dummy', max_length=128, unique=True, editable=False)

            class Fields:
                code = 'Non-secret code'
                expired = {'help_text': 'Do not check it. Do not.'}

    """


def get_model_class_from_string(model_path: str) -> Type[Model]:
    """Returns a certain model as defined in a string formatted `<app_name>.<model_name>`.

    Example:

        model = get_model_class_from_string('myapp.MyModel')

    """
    try:
        app_name, model_name = model_path.split('.')

    except ValueError:
        raise ImproperlyConfigured(
            f'`{model_path}` must have the following format: `app_name.model_name`.')

    try:
        model = apps.get_model(app_name, model_name)

    except (LookupError, ValueError):
        model = None

    if model is None:
        raise ImproperlyConfigured(
            f'`{model_path}` refers to a model `{model_name}` that has not been installed.')

    return model


def get_model_class_from_settings(settings_module: ModuleType, settings_entry_name: str):
    """Returns a certain model as defined in a given settings module.

    Example:

        myapp/settings.py

            from django.conf import settings

            MY_MODEL = getattr(settings, 'MYAPP_MY_MODEL', 'myapp.MyModel')


        myapp/some.py

            from myapp import settings

            model = get_model_class_from_settings(settings, 'MY_MODEL')

    """
    return get_model_class_from_string(getattr(settings_module, settings_entry_name))





def test_custom_model_pages(request_client, user_create):
    user = user_create(superuser=True)
    client = request_client(user=user)

    response = client.get('/admin/admin/')
    assert '/admin/admin/mypage2/' in response.rendered_content
    assert '/admin/admin/mypage1/' in response.rendered_content

    response = client.get('/admin/admin/mypage2/')
    assert 'Put data here' in response.rendered_content

    response = client.post('/admin/admin/mypage2/', data={'my_another_field': 'hohoho'}, follow=True)
    assert 'Test2:hohoho' in response.rendered_content

    response = client.post('/admin/admin/mypage1/', data={'my_field': '1234'}, follow=True)
    assert 'Test1:1234' in response.rendered_content
    assert 'Done.' in response.rendered_content



from typing import Union, Set, List, Tuple, Dict, Type, Any


class ChoicesEnumMixin:
    """Mixin to add to your enum classes.
    Allows custom titles and hint for enum elements.

    An enum with this mixin has:
        * titles
        * hints

    Every enum item has:
        * title
        * hint

    To be used in conjunction with `get_choices()`.

    .. code-block:: python

        from enum import Enum, unique

        @unique
        class Role(ChoicesEnumMixin, Enum):

            APPLICANT = 0, 'Title', 'Hint'
            ADMIN = 1, 'Administrator'
            MEMBER = 2

        class MyChoiceModel(models.Model):

            role = models.PositiveIntegerField(
                choices=get_choices(Role), default=Role.MEMBER)

        members = MyChoiceModel.objects.filter(role=Role.MEMBER)

    """
    titles = None
    """Value to title mapping."""

    hints = None
    """Value to hint mapping."""

    def __init__(self, *args):

        cls = self.__class__
        value = self.value

        def contribute_to_mapping(name, val):

            mapping = getattr(cls, name, None)

            if mapping is None:
                mapping = {}
                setattr(cls, name, mapping)

            mapping[value] = val

        title = self.title

        if title is None:
            title = self.name.lower().capitalize()
            self.title = title

        contribute_to_mapping('titles', title)
        contribute_to_mapping('hints', self.hint)

    def __new__(cls, *value):
        val = value[0]

        obj = object.__new__(cls)
        obj._value_ = val

        title = None
        hint = ''

        meta = value[1:]

        if meta:
            title = meta[0]

            if len(meta) > 1:
                hint = meta[1]

        obj.title = title
        obj.hint = hint

        return obj

    def __bool__(self):
        # Try to be compatible with boolean fields and lookups.
        return bool(self.value)

    def __str__(self):
        # Try to be compatible with text fields and lookups.
        return str(self.value)

    def __int__(self):
        # Try to be compatible with integer fields and lookups.
        return int(self.value)

    @classmethod
    def get_title(cls, item) -> str:
        """Returns a title for a variant instance or it's value.

        :param item:

        """
        if isinstance(item, cls):
            return item.title

        return cls.titles[item]

    @classmethod
    def get_hint(cls, item) -> str:
        """Returns a hint for a variant instance or it's value.

        :param item:

        """
        if isinstance(item, cls):
            return item.hint

        return cls.hints[item]


def choices_list(*choices: Union[Set, List, Tuple, Dict]) -> dict:
    """Helps to define choices for models, that could be addressed
    later as dictionaries.

    To be used in conjunction with `get_choices()`.

    Returns choices ordered dictionary.

    .. code-block:: python

        class MyModel(models.Model):

            TYPE_ONE = 1
            TYPE_TWO = 2

            TYPES = choices_list(
                (TYPE_ONE, 'Type one title'),
                (TYPE_TWO, 'Type two title'),
            )

            type = models.PositiveIntegerField('My type', choices=get_choices(TYPES), default=TYPE_TWO)

            def get_display_type(self):
                return self.TYPES[self.type]

    :param choices:

    """
    return dict(choices)


def get_choices(choices_list: Union[Dict, Type[ChoicesEnumMixin]]) -> Tuple[Tuple[Any, Any], ...]:
    """Returns model field choices from a given choices list.

    :param  choices_list:
        The list can be defined with `choices_list()` or could be an `ChoicesEnumMixin` and `Enum` subclass.

    """
    if isinstance(choices_list, type) and issubclass(choices_list, ChoicesEnumMixin):
        return tuple(choices_list.titles.items())

    return tuple((key, val) for key, val in choices_list.items())



from typing import Optional

from django.contrib.admin.decorators import register
from django.db import models
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from .admins import CustomPageModelAdmin, EtcAdmin


class CustomModelPage(models.Model):
    """Allows construction of admin pages based on user input.

    Define your fields (as usual in models) and override .save() method.

    .. code-block:: python

        class MyPage(CustomModelPage):

            title = 'Test page 1'  # set page title

            # Define some fields.
            my_field = models.CharField('some title', max_length=10)

            def save(self):
                ...  # Implement data handling.
                super().save()

        # Register my page within Django admin.
        MyPage.register()

    """
    title: str = _('Custom page')
    """Page title to be used."""

    app_label: str = 'admin'
    """Application label to relate page to. Default: admin"""

    bound_request: Optional[HttpRequest] = None
    """Request object bound to the model"""

    bound_admin: Optional[EtcAdmin] = None
    """Django admin model bound to this model."""

    class Meta:
        abstract = True
        managed = False

    @classmethod
    def __init_subclass__(cls) -> None:
        meta = cls.Meta
        meta.verbose_name = meta.verbose_name_plural = cls.title
        meta.app_label = cls.app_label
        super().__init_subclass__()

    @classmethod
    def register(cls, *, admin_model: CustomPageModelAdmin = None):
        """Registers this model page class in Django admin.

        :param admin_model:

        """
        register(cls)(admin_model or CustomPageModelAdmin)

    def save(self):  # noqa
        """Heirs should implement their own save handling."""
        self.bound_admin.message_success(self.bound_request, _('Done.'))



from pytest_djangoapp import configure_djangoapp_plugin


pytest_plugins = configure_djangoapp_plugin(
    {
        'SITE_ID': 1,
        'INSTALLED_APPS': [
            'django.contrib.sites',
            'django.contrib.contenttypes',
            'django.contrib.auth',
        ],
    },
    extend_MIDDLEWARE=[
        # messages in admin support
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
    ],
    admin_contrib=True,
)



from django.contrib import admin
from django.urls import path

from pytest_djangoapp.compat import get_urlpatterns

from .views import index


urlpatterns = get_urlpatterns([
    path('index/', index, name='index'),
    path('admin/', admin.site.urls),
])



from django import template

register = template.Library()


@register.simple_tag
def model_meta_verbose_name(model):
    """Returns model verbose name singular.

    Example:

        {% load model_meta %}
        {% model_meta_verbose_name my_model %}

    :param Model model:
    """
    return model._meta.verbose_name


@register.simple_tag
def model_meta_verbose_name_plural(model):
    """Returns model verbose name plural.

    Example:

        {% load model_meta %}
        {% model_meta_verbose_name_plural my_model %}

    :param Model model:
    :return:
    """
    return model._meta.verbose_name_plural
